from edc_constants.constants import OTHER

list_data = {
    "edc_mnsi.abnormalfootappearanceobservations": [
        ("callous_formation", "Callous formation"),
        ("deformity_amputation", "Deformity – amputation"),
        ("deformity_flat_feet", "Deformity – flat feet"),
        ("deformity_halux_valgus", "Deformity – halux valgus"),
        ("deformity_hammer_toes", "Deformity – hammer toes"),
        ("deformity_joint_subluxation", "Deformity – joint subluxation"),
        ("deformity_medial_convexity", "Deformity – medial convexity (Charcot foot)"),
        ("deformity_overlapping_toes", "Deformity – overlapping toes"),
        ("deformity_prominent_metatarsal_heads", "Deformity – prominent metatarsal heads"),
        ("dry_skin", "Dry skin"),
        ("infection", "Infection"),
        ("fissure", "Fissure"),
        (OTHER, "Other abnormality, please specify"),
    ],
}



from django.contrib import admin

from .fieldsets import get_fieldsets

radio_fields = {
    "amputation": admin.VERTICAL,
    "ankle_reflexes_left_foot": admin.VERTICAL,
    "ankle_reflexes_right_foot": admin.VERTICAL,
    "burning_pain_legs_feet": admin.VERTICAL,
    "covers_touch_skin_painful": admin.VERTICAL,
    "diabetic_neuropathy": admin.VERTICAL,
    "differentiate_hot_cold_water": admin.VERTICAL,
    "feel_weak": admin.VERTICAL,
    "feet_sensitive_touch": admin.VERTICAL,
    "legs_hurt_when_walk": admin.VERTICAL,
    "mnsi_performed": admin.VERTICAL,
    "monofilament_left_foot": admin.VERTICAL,
    "monofilament_right_foot": admin.VERTICAL,
    "muscle_cramps_legs_feet": admin.VERTICAL,
    "normal_appearance_left_foot": admin.VERTICAL,
    "normal_appearance_right_foot": admin.VERTICAL,
    "numb_legs_feet": admin.VERTICAL,
    "open_sore_foot_history": admin.VERTICAL,
    "prickling_feelings_legs_feet": admin.VERTICAL,
    "sense_feet_when_walk": admin.VERTICAL,
    "skin_cracks_open_feet": admin.VERTICAL,
    "symptoms_worse_night": admin.VERTICAL,
    "ulceration_left_foot": admin.VERTICAL,
    "ulceration_right_foot": admin.VERTICAL,
    "vibration_perception_left_toe": admin.VERTICAL,
    "vibration_perception_right_toe": admin.VERTICAL,
}


class MnsiModelAdminMixin:
    form = None

    fieldsets = get_fieldsets()

    filter_horizontal = (
        "abnormal_obs_left_foot",
        "abnormal_obs_right_foot",
    )

    readonly_fields = (
        "calculated_patient_history_score",
        "calculated_physical_assessment_score",
    )

    radio_fields = radio_fields



from .mnsi_model_mixin import MnsiModelMixin
from .model_fields_mixins import (
    MnsiFieldsModelMixin,
    MnsiLeftFootFieldsModelMixin,
    MnsiRightFootFieldsModelMixin,
)
from .model_methods_mixins import MnsiMethodsModelMixin



# Generated by Django 4.0.4 on 2022-05-12 23:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_mnsi", "0002_historicalmnsi_mnsi"),
    ]

    operations = [
        migrations.AlterField(
            model_name="mnsi",
            name="abnormal_obs_left_foot",
            field=models.ManyToManyField(
                blank=True,
                related_name="+",
                to="edc_mnsi.abnormalfootappearanceobservations",
                verbose_name="If NO, check all that apply to LEFT foot?",
            ),
        ),
        migrations.AlterField(
            model_name="mnsi",
            name="abnormal_obs_right_foot",
            field=models.ManyToManyField(
                blank=True,
                related_name="+",
                to="edc_mnsi.abnormalfootappearanceobservations",
                verbose_name="If NO, check all that apply to RIGHT foot?",
            ),
        ),
    ]



from django.test import TestCase
from edc_constants.constants import (
    ABSENT,
    DECREASED,
    NO,
    NOT_EXAMINED,
    PRESENT,
    PRESENT_WITH_REINFORCEMENT,
    REDUCED,
    YES,
)

from edc_mnsi.calculator import (
    MnsiCalculator,
    MnsiPatientHistoryCalculatorError,
    MnsiPhysicalAssessmentCalculatorError,
)

from .mixins import TestCaseMixin


class TestMnsiCalculators(TestCaseMixin, TestCase):
    def test_calculator_instantiated_with_dict(self):
        responses = self.get_best_case_answers()
        mnsi_calculator = MnsiCalculator(**responses)
        self.assertEqual(mnsi_calculator.patient_history_score(), 0)
        self.assertEqual(mnsi_calculator.physical_assessment_score(), 0)

    def test_calculator_instantiated_with_model(self):
        responses = self.get_best_case_answers()
        model = self.get_mnsi_obj(**responses)
        mnsi_calculator = MnsiCalculator(model)
        self.assertEqual(mnsi_calculator.patient_history_score(), 0)
        self.assertEqual(mnsi_calculator.physical_assessment_score(), 0)

    def test_calculator_instantiated_with_dict2(self):
        responses = self.get_best_case_answers()
        responses.update(self.get_worst_case_patient_history_data())
        responses.update(self.get_worst_case_physical_assessment_data())
        mnsi_calculator = MnsiCalculator(**responses)
        self.assertEqual(mnsi_calculator.patient_history_score(), 13)
        self.assertEqual(mnsi_calculator.physical_assessment_score(), 10)

    def test_calculator_instantiated_with_model2(self):
        responses = self.get_best_case_answers()
        responses.update(self.get_worst_case_patient_history_data())
        responses.update(self.get_worst_case_physical_assessment_data())
        model = self.get_mnsi_obj(**responses)
        mnsi_calculator = MnsiCalculator(model)
        self.assertEqual(mnsi_calculator.patient_history_score(), 13)
        self.assertEqual(mnsi_calculator.physical_assessment_score(), 10)

    def test_calculator_instantiated_with_dict_when_mnsi_not_performed(self):
        responses = self.get_mnsi_not_performed_answers()
        mnsi_calculator = MnsiCalculator(**responses)
        self.assertEqual(mnsi_calculator.patient_history_score(), 0)
        self.assertEqual(mnsi_calculator.physical_assessment_score(), 0)

    def test_calculator_instantiated_with_model_when_mnsi_not_performed(self):
        responses = self.get_mnsi_not_performed_answers()
        model = self.get_mnsi_obj(**responses)
        mnsi_calculator = MnsiCalculator(model)
        self.assertEqual(mnsi_calculator.patient_history_score(), 0)
        self.assertEqual(mnsi_calculator.physical_assessment_score(), 0)

    def test_missing_required_field_raises_mnsi_patient_history_calculator_error(
        self,
    ):
        responses = self.get_best_case_answers()
        responses.pop("amputation")
        mnsi_calculator = MnsiCalculator(**responses)
        with self.assertRaises(MnsiPatientHistoryCalculatorError):
            mnsi_calculator.patient_history_score()

    def test_missing_non_required_fields_does_not_raise_mnsi_patient_history_calculator_error(
        self,
    ):
        responses = self.get_best_case_answers()
        responses.pop("muscle_cramps_legs_feet")
        responses.pop("feel_weak")
        mnsi_calculator = MnsiCalculator(**responses)
        try:
            mnsi_calculator.patient_history_score()
        except MnsiPatientHistoryCalculatorError as exc:
            self.fail(
                f"mnsi_calculator.patient_history_score() raised "
                f"MnsiPatientHistoryCalculatorError unexpectedly.\nDetails: {exc}"
            )

    def test_missing_required_field_raises_mnsi_physical_assessment_calculator_error(
        self,
    ):
        responses = self.get_best_case_answers()
        responses.pop("ulceration_left_foot")
        mnsi_calculator = MnsiCalculator(**responses)
        with self.assertRaises(MnsiPhysicalAssessmentCalculatorError):
            mnsi_calculator.physical_assessment_score()

    def test_best_case_patient_history_returns_min_score_of_zero(self):
        mnsi_calculator = MnsiCalculator(**self.get_best_case_answers())
        self.assertEqual(mnsi_calculator.patient_history_score(), 0)

    def test_patient_history_returns_score_of_zero_when_mnsi_not_performed(self):
        mnsi_calculator = MnsiCalculator(**self.get_mnsi_not_performed_answers())
        self.assertEqual(mnsi_calculator.patient_history_score(), 0)

    def test_worst_case_patient_history_returns_max_score_of_thirteen(self):
        responses = self.get_best_case_answers()
        responses.update(self.get_worst_case_patient_history_data())
        mnsi_calculator = MnsiCalculator(**responses)
        self.assertEqual(mnsi_calculator.patient_history_score(), 13)

    def test_q4_and_q10_do_not_affect_patient_history_score(self):
        # Best case score should be 0
        responses = self.get_best_case_answers()
        mnsi_calculator = MnsiCalculator(**responses)
        self.assertEqual(mnsi_calculator.patient_history_score(), 0)

        # Best case score should remain 0 after modifying q4 and 10
        responses.update({"muscle_cramps_legs_feet": YES, "feel_weak": YES})
        mnsi_calculator = MnsiCalculator(**responses)
        self.assertEqual(mnsi_calculator.patient_history_score(), 0)

        # Worst case score should be 13
        responses.update(self.get_worst_case_patient_history_data())
        mnsi_calculator = MnsiCalculator(**responses)
        self.assertEqual(mnsi_calculator.patient_history_score(), 13)

        # Best case score should remain 13 after modifying q4 and 10
        responses.update({"muscle_cramps_legs_feet": NO, "feel_weak": NO})
        mnsi_calculator = MnsiCalculator(**responses)
        self.assertEqual(mnsi_calculator.patient_history_score(), 13)

    def test_best_case_physical_assessment_returns_min_score_of_zero(self):
        mnsi_calculator = MnsiCalculator(**self.get_best_case_answers())
        self.assertEqual(mnsi_calculator.physical_assessment_score(), 0)

    def test_physical_assessment_returns_score_of_zero_when_mnsi_not_performed(self):
        mnsi_calculator = MnsiCalculator(**self.get_mnsi_not_performed_answers())
        self.assertEqual(mnsi_calculator.physical_assessment_score(), 0)

    def test_worst_case_physical_assessment_returns_max_score_of_ten(self):
        responses = self.get_best_case_answers()
        responses.update(self.get_worst_case_physical_assessment_data())
        mnsi_calculator = MnsiCalculator(**responses)
        self.assertEqual(mnsi_calculator.physical_assessment_score(), 10)

    def test_patient_history_scores_where_YES_awards_one_point(self):
        one_point_if_yes_response_questions = [
            "numb_legs_feet",  # Q1
            "burning_pain_legs_feet",  # Q2
            "feet_sensitive_touch",  # Q3
            "prickling_feelings_legs_feet",  # Q5
            "covers_touch_skin_painful",  # Q6
            "open_sore_foot_history",  # Q8
            "diabetic_neuropathy",  # Q9
            "symptoms_worse_night",  # Q11
            "legs_hurt_when_walk",  # Q12
            "skin_cracks_open_feet",  # Q14
            "amputation",  # Q15
        ]

        for question in one_point_if_yes_response_questions:
            with self.subTest(
                f"Testing '{question}' with 'YES' response is worth 1 point",
                question=question,
            ):
                responses = self.get_best_case_answers()
                responses[question] = YES
                mnsi_calculator = MnsiCalculator(**responses)
                self.assertEqual(mnsi_calculator.patient_history_score(), 1)

    def test_patient_history_scores_where_NO_awards_one_point(self):
        one_point_if_no_response_questions = [
            "differentiate_hot_cold_water",  # Q7
            "sense_feet_when_walk",  # Q13
        ]

        for question in one_point_if_no_response_questions:
            with self.subTest(
                f"Testing '{question}' with 'NO' response is worth 1 point",
                question=question,
            ):
                responses = self.get_best_case_answers()
                responses[question] = NO
                mnsi_calculator = MnsiCalculator(**responses)
                self.assertEqual(mnsi_calculator.patient_history_score(), 1)

    def test_physical_assessment_abnormal_foot_appearance_awards_one_point(self):
        normal_foot_appearance_questions = [
            "normal_appearance_right_foot",
            "normal_appearance_left_foot",
        ]

        for question in normal_foot_appearance_questions:
            with self.subTest(
                f"Testing '{question}' with 'NO' response is worth 1 point",
                question=question,
            ):
                responses = self.get_best_case_answers()
                responses[question] = NO
                mnsi_calculator = MnsiCalculator(**responses)
                self.assertEqual(mnsi_calculator.physical_assessment_score(), 1.0)

    def test_physical_assessment_amputation_with_no_further_examinations_awards_one_point(
        self,
    ):
        for foot_choice in self.foot_choices:
            for amputated_obs_set in self.amputated_abnormal_obs_sets:
                with self.subTest(
                    (
                        f"Testing '{foot_choice}' amputated with no further "
                        "examinations is worth 1 point"
                    ),
                    foot_choice=foot_choice,
                    abnormal_obs=amputated_obs_set,
                ):
                    responses = self.get_best_case_answers()
                    responses.update(
                        {
                            # Specify foot abnormal appearance, amputated
                            f"normal_appearance_{foot_choice}_foot": NO,
                            f"abnormal_obs_{foot_choice}_foot": amputated_obs_set,
                            # Skip other physical assessment answers
                            f"ulceration_{foot_choice}_foot": NOT_EXAMINED,
                            f"ankle_reflexes_{foot_choice}_foot": NOT_EXAMINED,
                            f"vibration_perception_{foot_choice}_toe": NOT_EXAMINED,
                            f"monofilament_{foot_choice}_foot": NOT_EXAMINED,
                        }
                    )

                    mnsi_calculator = MnsiCalculator(**responses)
                    self.assertEqual(mnsi_calculator.physical_assessment_score(), 1.0)

    def test_physical_assessment_foot_ulceration_present_awards_one_point(self):
        ulceration_questions = [
            "ulceration_right_foot",
            "ulceration_left_foot",
        ]

        for question in ulceration_questions:
            with self.subTest(
                f"Testing '{question}' with 'PRESENT' response is worth 1 point",
                question=question,
            ):
                responses = self.get_best_case_answers()
                responses[question] = PRESENT
                mnsi_calculator = MnsiCalculator(**responses)
                self.assertEqual(mnsi_calculator.physical_assessment_score(), 1.0)

    def test_physical_assessment_ankle_reflexes_present_reinforcement_awards_half_point(
        self,
    ):
        ankle_reflex_questions = [
            "ankle_reflexes_right_foot",
            "ankle_reflexes_left_foot",
        ]

        for question in ankle_reflex_questions:
            with self.subTest(
                f"Testing '{question}' with 'PRESENT_REINFORCEMENT' response "
                "is worth 0.5 point",
                question=question,
            ):
                responses = self.get_best_case_answers()
                responses[question] = PRESENT_WITH_REINFORCEMENT
                mnsi_calculator = MnsiCalculator(**responses)
                self.assertEqual(mnsi_calculator.physical_assessment_score(), 0.5)

    def test_physical_assessment_ankle_reflexes_absent_awards_one_point(
        self,
    ):
        ankle_reflex_questions = [
            "ankle_reflexes_right_foot",
            "ankle_reflexes_left_foot",
        ]

        for question in ankle_reflex_questions:
            with self.subTest(
                f"Testing '{question}' with 'ABSENT' response is worth 1 point",
                question=question,
            ):
                responses = self.get_best_case_answers()
                responses[question] = ABSENT
                mnsi_calculator = MnsiCalculator(**responses)
                self.assertEqual(mnsi_calculator.physical_assessment_score(), 1)

    def test_physical_assessment_vibration_perception_decreased_awards_half_point(
        self,
    ):
        vibration_perception_questions = [
            "vibration_perception_right_toe",
            "vibration_perception_left_toe",
        ]

        for question in vibration_perception_questions:
            with self.subTest(
                f"Testing '{question}' with 'DECREASED' response is worth 0.5 point",
                question=question,
            ):
                responses = self.get_best_case_answers()
                responses[question] = DECREASED
                mnsi_calculator = MnsiCalculator(**responses)
                self.assertEqual(mnsi_calculator.physical_assessment_score(), 0.5)

    def test_physical_assessment_vibration_perception_absent_awards_one_point(
        self,
    ):
        vibration_perception_questions = [
            "vibration_perception_right_toe",
            "vibration_perception_left_toe",
        ]

        for question in vibration_perception_questions:
            with self.subTest(
                f"Testing '{question}' with 'ABSENT' response is worth 1 point",
                question=question,
            ):
                responses = self.get_best_case_answers()
                responses[question] = ABSENT
                mnsi_calculator = MnsiCalculator(**responses)
                self.assertEqual(mnsi_calculator.physical_assessment_score(), 1)

    def test_physical_assessment_monofilament_reduced_awards_half_point(
        self,
    ):
        monofilament_questions = [
            "monofilament_right_foot",
            "monofilament_left_foot",
        ]

        for question in monofilament_questions:
            with self.subTest(
                f"Testing '{question}' with 'REDUCED' response is worth 0.5 point",
                question=question,
            ):
                responses = self.get_best_case_answers()
                responses[question] = REDUCED
                mnsi_calculator = MnsiCalculator(**responses)
                self.assertEqual(mnsi_calculator.physical_assessment_score(), 0.5)

    def test_physical_assessment_monofilament_absent_awards_one_point(
        self,
    ):
        monofilament_questions = [
            "monofilament_right_foot",
            "monofilament_left_foot",
        ]

        for question in monofilament_questions:
            with self.subTest(
                f"Testing '{question}' with 'ABSENT' response is worth 0.5 point",
                question=question,
            ):
                responses = self.get_best_case_answers()
                responses[question] = ABSENT
                mnsi_calculator = MnsiCalculator(**responses)
                self.assertEqual(mnsi_calculator.physical_assessment_score(), 1)



calculated_values_fieldset = (
    "Calculated values",
    {
        "classes": ("collapse",),
        "fields": (
            "calculated_patient_history_score",
            "calculated_physical_assessment_score",
        ),
    },
)

patient_history_fields = (
    "numb_legs_feet",
    "burning_pain_legs_feet",
    "feet_sensitive_touch",
    "muscle_cramps_legs_feet",
    "prickling_feelings_legs_feet",
    "covers_touch_skin_painful",
    "differentiate_hot_cold_water",
    "open_sore_foot_history",
    "diabetic_neuropathy",
    "feel_weak",
    "symptoms_worse_night",
    "legs_hurt_when_walk",
    "sense_feet_when_walk",
    "skin_cracks_open_feet",
    "amputation",
)


def get_physical_assessment_fields(foot_choice):
    return (
        f"normal_appearance_{foot_choice}_foot",
        f"abnormal_obs_{foot_choice}_foot",
        f"abnormal_obs_{foot_choice}_foot_other",
        f"ulceration_{foot_choice}_foot",
        f"ankle_reflexes_{foot_choice}_foot",
        f"vibration_perception_{foot_choice}_toe",
        f"monofilament_{foot_choice}_foot",
    )


def get_fieldsets():
    return (
        (
            "Part 1: Patient History",
            {
                "description": (
                    "To be completed by the patient. If the MNSI assessment "
                    "was not performed, the response is `not applicable`."
                ),
                "fields": patient_history_fields,
            },
        ),
        (
            "Part 2a: Physical Assessment - Right Foot",
            {
                "description": "To be completed by health professional",
                "fields": get_physical_assessment_fields("right"),
            },
        ),
        (
            "Part 2b: Physical Assessment - Left Foot",
            {
                "description": "To be completed by health professional",
                "fields": get_physical_assessment_fields("left"),
            },
        ),
    )



from copy import deepcopy

from django.test import TestCase
from edc_constants.constants import NO, NORMAL, NOT_EXAMINED, OTHER, PRESENT, YES
from edc_form_validators import FormValidatorTestCaseMixin

from edc_mnsi.calculator import (
    MnsiPatientHistoryCalculatorError,
    MnsiPhysicalAssessmentCalculatorError,
)
from edc_mnsi.models import AbnormalFootAppearanceObservations, Mnsi

from ..forms import MnsiForm, MnsiFormValidator
from .mixins import TestCaseMixin


class TestMnsiFormValidator(FormValidatorTestCaseMixin, TestCaseMixin, TestCase):
    form_validator_cls = MnsiFormValidator

    def test_valid_best_case_form_ok(self):
        cleaned_data = deepcopy(self.get_best_case_answers())
        form = MnsiForm(data=cleaned_data)
        form.is_valid()
        self.assertEqual(form._errors, {})

    def test_valid_mnsi_not_performed_form_ok(self):
        cleaned_data = deepcopy(self.get_mnsi_not_performed_answers())
        form = MnsiForm(data=cleaned_data)
        form.is_valid()
        self.assertEqual(form._errors, {})

    def test_physical_assessment_questions_not_applicable_if_mnsi_not_performed(self):
        for foot_choice in self.foot_choices:
            for question_field in [
                f"normal_appearance_{foot_choice}_foot",
                f"ulceration_{foot_choice}_foot",
                f"ankle_reflexes_{foot_choice}_foot",
                f"vibration_perception_{foot_choice}_toe",
                f"monofilament_{foot_choice}_foot",
            ]:
                # Setup test case
                cleaned_data = deepcopy(self.get_mnsi_not_performed_answers())
                # set one field as answered, e.g. != NOT_EXAMINED
                cleaned_data.update(
                    {
                        question_field: self.get_best_case_answers()[question_field],
                    }
                )

                # Test
                with self.subTest(foot_choice=foot_choice, question_field=question_field):
                    form_validator = self.validate_form_validator(cleaned_data)
                    self.assertIn(question_field, form_validator._errors)
                    self.assertIn(
                        "Invalid. "
                        "Expected `not examined` if MNSI assessment was not performed.",
                        str(form_validator._errors.get(question_field)),
                    )
                    self.assertEqual(len(form_validator._errors), 1, form_validator._errors)

    def test_valid_worst_case_form_ok(self):
        cleaned_data = deepcopy(self.get_best_case_answers())
        cleaned_data.update(self.get_worst_case_patient_history_data())
        cleaned_data.update(self.get_worst_case_physical_assessment_data())
        form = MnsiForm(data=cleaned_data)
        form.is_valid()
        self.assertEqual(form._errors, {})

    def test_foot_amputated_returns_false_if_foot_normal_appearance(self):
        for foot_choice in self.foot_choices:
            with self.subTest(foot_choice=foot_choice):
                cleaned_data = deepcopy(self.get_best_case_answers())
                cleaned_data.update(
                    {
                        f"normal_appearance_{foot_choice}_foot": YES,
                        f"abnormal_obs_{foot_choice}_foot": self.get_nonempty_abnormal_obs_set(
                            "deformity_amputation"
                        ),
                    }
                )
                form_validator = self.form_validator_cls(cleaned_data=cleaned_data, model=Mnsi)
                self.assertFalse(form_validator.foot_amputated(foot_choice))

    def test_foot_amputated_returns_false_if_no_abnormalities_on_foot(self):
        for foot_choice in self.foot_choices:
            with self.subTest(foot_choice=foot_choice):
                cleaned_data = deepcopy(self.get_best_case_answers())
                cleaned_data.update(
                    {
                        f"normal_appearance_{foot_choice}_foot": NO,
                        f"abnormal_obs_{foot_choice}_foot": self.get_empty_abnormal_obs_set(),
                    }
                )
                form_validator = self.form_validator_cls(cleaned_data=cleaned_data, model=Mnsi)
                self.assertFalse(form_validator.foot_amputated(foot_choice))

    def test_foot_amputated_returns_false_if_abnormality_not_amputated(self):
        for foot_choice in self.foot_choices:
            with self.subTest(foot_choice=foot_choice):
                cleaned_data = deepcopy(self.get_best_case_answers())
                cleaned_data.update(
                    {
                        f"normal_appearance_{foot_choice}_foot": NO,
                        f"abnormal_obs_{foot_choice}_foot": self.get_nonempty_abnormal_obs_set(
                            "infection"
                        ),
                    }
                )
                form_validator = self.form_validator_cls(cleaned_data=cleaned_data, model=Mnsi)
                self.assertFalse(form_validator.foot_amputated(foot_choice))

    def test_foot_amputated_returns_true_if_foot_amputated(self):
        for foot_choice in self.foot_choices:
            for amputated_obs_set in self.amputated_abnormal_obs_sets:
                with self.subTest(
                    foot_choice=foot_choice,
                    abnormal_obs=amputated_obs_set,
                ):
                    cleaned_data = {
                        f"normal_appearance_{foot_choice}_foot": NO,
                        f"abnormal_obs_{foot_choice}_foot": amputated_obs_set,
                    }
                    form_validator = self.form_validator_cls(
                        cleaned_data=cleaned_data, model=Mnsi
                    )
                    self.assertTrue(form_validator.foot_amputated(foot_choice))

    def test_physical_assessment_questions_applicable_if_foot_not_amputated(self):
        for foot_choice in self.foot_choices:
            for question_field in [
                f"ulceration_{foot_choice}_foot",
                f"ankle_reflexes_{foot_choice}_foot",
                f"vibration_perception_{foot_choice}_toe",
                f"monofilament_{foot_choice}_foot",
            ]:
                # Setup test case
                cleaned_data = self.get_best_case_answers()
                cleaned_data.update({question_field: NOT_EXAMINED})

                # Test
                with self.subTest(foot_choice=foot_choice, question_field=question_field):
                    form_validator = self.validate_form_validator(cleaned_data)
                    self.assertIn(question_field, form_validator._errors)
                    self.assertIn(
                        "Invalid. "
                        "Examination result expected if MNSI assessment was performed.",
                        str(form_validator._errors.get(question_field)),
                    )
                    self.assertEqual(len(form_validator._errors), 1, form_validator._errors)

    def test_physical_assessment_questions_applicable_if_foot_abnormality_but_not_amputated(
        self,
    ):
        for foot_choice in self.foot_choices:
            for question_field in [
                f"ulceration_{foot_choice}_foot",
                f"ankle_reflexes_{foot_choice}_foot",
                f"vibration_perception_{foot_choice}_toe",
                f"monofilament_{foot_choice}_foot",
            ]:
                # Setup test case
                cleaned_data = self.get_best_case_answers()
                cleaned_data.update(
                    {
                        # Specify foot abnormal appearance, not-amputated
                        f"normal_appearance_{foot_choice}_foot": NO,
                        f"abnormal_obs_{foot_choice}_foot": self.get_nonempty_abnormal_obs_set(
                            "infection"
                        ),
                    }
                )
                cleaned_data.update({question_field: NOT_EXAMINED})

                # Test
                with self.subTest(foot_choice=foot_choice, question_field=question_field):
                    form_validator = self.validate_form_validator(cleaned_data)
                    self.assertIn(question_field, form_validator._errors)
                    self.assertIn(
                        "Invalid. "
                        "Examination result expected if MNSI assessment was performed.",
                        str(form_validator._errors.get(question_field)),
                    )
                    self.assertEqual(len(form_validator._errors), 1, form_validator._errors)

    def test_physical_assessment_questions_not_required_if_foot_amputated(self):
        for foot_choice in self.foot_choices:
            for amputated_obs_set in self.amputated_abnormal_obs_sets:
                with self.subTest(foot_choice=foot_choice, abnormal_obs=amputated_obs_set):
                    cleaned_data = deepcopy(self.get_best_case_answers())
                    cleaned_data.update(
                        {
                            # Specify foot abnormal appearance, amputated
                            f"normal_appearance_{foot_choice}_foot": NO,
                            f"abnormal_obs_{foot_choice}_foot": amputated_obs_set,
                            # Skip other physical assessment answers
                            f"ulceration_{foot_choice}_foot": NOT_EXAMINED,
                            f"ankle_reflexes_{foot_choice}_foot": NOT_EXAMINED,
                            f"vibration_perception_{foot_choice}_toe": NOT_EXAMINED,
                            f"monofilament_{foot_choice}_foot": NOT_EXAMINED,
                        }
                    )

                    form_validator = self.validate_form_validator(cleaned_data)
                    self.assertEqual(form_validator._errors, {})

    def test_physical_assessment_questions_still_allowed_if_foot_amputated(self):
        for foot_choice in self.foot_choices:
            for amputated_obs_set in self.amputated_abnormal_obs_sets:
                with self.subTest(foot_choice=foot_choice, abnormal_obs=amputated_obs_set):
                    cleaned_data = self.get_best_case_answers()
                    cleaned_data.update(
                        {
                            # Specify foot abnormal appearance, amputated
                            f"normal_appearance_{foot_choice}_foot": NO,
                            f"abnormal_obs_{foot_choice}_foot": amputated_obs_set,
                            # Skip other physical assessment answers
                            f"ulceration_{foot_choice}_foot": PRESENT,
                            f"ankle_reflexes_{foot_choice}_foot": PRESENT,
                            f"vibration_perception_{foot_choice}_toe": PRESENT,
                            f"monofilament_{foot_choice}_foot": NORMAL,
                        }
                    )

                    form_validator = self.validate_form_validator(cleaned_data)
                    self.assertEqual(form_validator._errors, {})

    def test_abnormal_observations_required_if_foot_appearance_not_normal(self):
        cleaned_data = deepcopy(self.get_best_case_answers())

        for foot_choice in self.foot_choices:
            field = f"normal_appearance_{foot_choice}_foot"
            m2m_field = f"abnormal_obs_{foot_choice}_foot"

            with self.subTest(
                f"Testing '{m2m_field}' is required if {field}='No'",
                field=field,
                m2m_field=m2m_field,
            ):
                cleaned_data.update({field: NO})
                form_validator = self.validate_form_validator(cleaned_data)
                self.assertIn(m2m_field, form_validator._errors)
                self.assertIn(
                    "This field is required",
                    str(form_validator._errors.get(m2m_field)),
                )
                self.assertEqual(len(form_validator._errors), 1, form_validator._errors)

                # Set back to YES, and move on to next test
                cleaned_data.update({field: YES})

    def test_abnormal_observations_accepted_if_foot_appearance_not_normal(self):
        cleaned_data = deepcopy(self.get_best_case_answers())
        m2m_field_selection = AbnormalFootAppearanceObservations.objects.filter(
            name="infection"
        )

        for foot_choice in self.foot_choices:
            field = f"normal_appearance_{foot_choice}_foot"
            m2m_field = f"abnormal_obs_{foot_choice}_foot"

            with self.subTest(
                f"Testing '{m2m_field}' accepted if {field}='No'",
                field=field,
                m2m_field=m2m_field,
            ):
                cleaned_data.update({field: NO, m2m_field: m2m_field_selection})
                form_validator = self.validate_form_validator(cleaned_data)
                self.assertEqual(form_validator._errors, {})

    def test_abnormal_observations_not_applicable_if_foot_appearance_is_normal(self):
        cleaned_data = deepcopy(self.get_best_case_answers())
        m2m_field_selection = AbnormalFootAppearanceObservations.objects.filter(
            name="infection"
        )

        for foot_choice in self.foot_choices:
            field = f"normal_appearance_{foot_choice}_foot"
            m2m_field = f"abnormal_obs_{foot_choice}_foot"

            with self.subTest(
                f"Testing '{m2m_field}' accepted if {field}='No'",
                field=field,
                m2m_field=m2m_field,
            ):
                cleaned_data.update({field: YES, m2m_field: m2m_field_selection})
                form_validator = self.validate_form_validator(cleaned_data)
                self.assertIn(m2m_field, form_validator._errors)
                self.assertIn(
                    "This field is not required",
                    str(form_validator._errors.get(m2m_field)),
                )
                self.assertEqual(len(form_validator._errors), 1, form_validator._errors)

                # Set to No (i.e. make m2m applicable) and move onto next test
                cleaned_data.update({field: NO})

    def test_other_field_required_if_other_specified(self):
        cleaned_data = deepcopy(self.get_best_case_answers())
        other_observation = AbnormalFootAppearanceObservations.objects.filter(name=OTHER)

        for foot_choice in self.foot_choices:
            field = f"normal_appearance_{foot_choice}_foot"
            m2m_field = f"abnormal_obs_{foot_choice}_foot"
            m2m_field_other = f"{m2m_field}_other"

            with self.subTest(
                f"Testing '{m2m_field_other}' required if {m2m_field}={other_observation}",
                field=field,
                m2m_field=m2m_field,
                m2m_field_other=m2m_field_other,
            ):
                # Select 'other', then test it's required
                cleaned_data.update({field: NO, m2m_field: other_observation})
                form_validator = self.validate_form_validator(cleaned_data)
                self.assertIn(m2m_field_other, form_validator._errors)
                self.assertIn(
                    "This field is required",
                    str(form_validator._errors.get(m2m_field_other)),
                )
                self.assertEqual(len(form_validator._errors), 1, form_validator._errors)

                # Complete 'other' field, and move onto next test
                cleaned_data.update({m2m_field_other: "Some other value"})

    def test_other_field_not_required_if_other_not_specified(self):
        cleaned_data = self.get_best_case_answers()
        non_other_observation = AbnormalFootAppearanceObservations.objects.filter(
            name="infection"
        )

        for foot_choice in self.foot_choices:
            field = f"normal_appearance_{foot_choice}_foot"
            m2m_field = f"abnormal_obs_{foot_choice}_foot"
            m2m_field_other = f"{m2m_field}_other"

            with self.subTest(
                f"Testing '{m2m_field_other}' completed when not required",
                field=field,
                m2m_field=m2m_field,
                m2m_field_other=m2m_field_other,
            ):
                # Try with normal foot appearance
                cleaned_data.update({field: YES, m2m_field_other: "Some other value"})
                form_validator = self.validate_form_validator(cleaned_data)
                self.assertIn(m2m_field_other, form_validator._errors)
                self.assertIn(
                    "This field is not required",
                    str(form_validator._errors.get(m2m_field_other)),
                )
                self.assertEqual(len(form_validator._errors), 1, form_validator._errors)

                # Try with abnormal foot appearance, and non-"other" observation
                cleaned_data.update(
                    {
                        field: NO,
                        m2m_field: non_other_observation,
                        m2m_field_other: "Some other value",
                    }
                )
                form_validator = self.validate_form_validator(cleaned_data)
                self.assertIn(m2m_field_other, form_validator._errors)
                self.assertIn(
                    "This field is not required",
                    str(form_validator._errors.get(m2m_field_other)),
                )
                self.assertEqual(len(form_validator._errors), 1, form_validator._errors)

                # Remove 'other' field value, make valid and move onto next test
                del cleaned_data[m2m_field]
                del cleaned_data[m2m_field_other]
                cleaned_data.update({field: YES})

    def test_missing_required_field_raises_mnsi_patient_history_calculator_error(
        self,
    ):
        cleaned_data = self.get_best_case_answers()
        cleaned_data.pop("numb_legs_feet")

        with self.assertRaises(MnsiPatientHistoryCalculatorError):
            self.validate_form_validator(cleaned_data)

    def test_missing_required_field_raises_mnsi_physical_assessment_calculator_error(
        self,
    ):
        cleaned_data = self.get_best_case_answers()
        cleaned_data.pop("ulceration_left_foot")

        with self.assertRaises(MnsiPhysicalAssessmentCalculatorError):
            self.validate_form_validator(cleaned_data)



from django.db import models

from .model_fields_mixins import (
    MnsiFieldsModelMixin,
    MnsiLeftFootFieldsModelMixin,
    MnsiRightFootFieldsModelMixin,
)
from .model_methods_mixins import MnsiMethodsModelMixin


class MnsiModelMixin(
    MnsiRightFootFieldsModelMixin,
    MnsiLeftFootFieldsModelMixin,
    MnsiFieldsModelMixin,
    MnsiMethodsModelMixin,
    models.Model,
):

    """Neuropathy screening tool.

    Uses Michigan Neuropathy Screening Instrument (MNSI), see:
        https://pubmed.ncbi.nlm.nih.gov/7821168/
        https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3641573/ (omits monofilament testing)
        https://medicine.umich.edu/sites/default/files/downloads/MNSI_howto.pdf

    """

    class Meta:
        abstract = True
        verbose_name = "Michigan Neuropathy Screening Instrument (MNSI)"
        verbose_name_plural = "Michigan Neuropathy Screening Instrument (MNSI)"



from django.db.models.signals import post_save
from django.dispatch import receiver


@receiver(post_save, weak=False, dispatch_uid="calculate_mnsi_score_on_post_save")
def calculate_mnsi_score_on_post_save(sender, instance, raw, created, using, **kwargs):
    if not raw and not kwargs.get("update_fields"):
        try:
            instance.update_mnsi_calculated_fields()
        except AttributeError as e:
            if "update_mnsi_calculated_fields" not in str(e):
                raise



from edc_auth.site_auths import site_auths

from .auth_objects import MNSI, MNSI_SUPER, MNSI_VIEW, mnsi_codenames

site_auths.add_group(*mnsi_codenames, name=MNSI_VIEW, view_only=True)
site_auths.add_group(*mnsi_codenames, name=MNSI, no_delete=True)
site_auths.add_group(*mnsi_codenames, name=MNSI_SUPER)



from .utils import get_mnsi_model_name

MNSI = "MNSI"
MNSI_VIEW = "MNSI_VIEW"
MNSI_SUPER = "MNSI_SUPER"

mnsi_codenames = []

app_name, model_name = get_mnsi_model_name().split(".")
for prefix in ["add", "change", "view", "delete"]:
    mnsi_codenames.append(f"{app_name}.{prefix}_{model_name}")
mnsi_codenames.append(f"{app_name}.view_historical{model_name}")
mnsi_codenames.append("edc_mnsi.view_abnormalfootappearanceobservations")
mnsi_codenames.sort()



from django import forms
from edc_form_validators import FormValidator, FormValidatorMixin
from edc_sites.forms import SiteModelFormMixin

from edc_mnsi.form_validator import MnsiFormValidatorMixin
from edc_mnsi.models import Mnsi


class MnsiFormValidator(MnsiFormValidatorMixin, FormValidator):
    pass


class MnsiForm(SiteModelFormMixin, FormValidatorMixin, forms.ModelForm):
    form_validator_cls = MnsiFormValidator

    class Meta:
        model = Mnsi
        fields = "__all__"



from django.test import TestCase
from edc_constants.constants import (
    ABSENT,
    NO,
    NORMAL,
    NOT_APPLICABLE,
    NOT_EXAMINED,
    PRESENT,
    YES,
)

from edc_mnsi.models import Mnsi

from .mixins import TestCaseMixin


class TestMnsiModel(TestCaseMixin, TestCase):
    def test_mnsi_calculations_saved_with_best_case_responses(self):
        responses = self.get_best_case_answers()
        model = self.get_mnsi_obj(**responses)

        # Confirm some assumptions about model attributes
        # (both non-calculated and calculated)
        self.assertEqual(model.numb_legs_feet, NO)
        self.assertEqual(model.ankle_reflexes_right_foot, PRESENT)
        self.assertEqual(model.monofilament_left_foot, NORMAL)
        self.assertEqual(model.calculated_patient_history_score, 0)
        self.assertEqual(model.calculated_physical_assessment_score, 0)

        # Retrieve saved object, and confirm the same assumptions hold
        retr_obj = Mnsi.objects.get(id=model.id)
        self.assertEqual(retr_obj.numb_legs_feet, NO)
        self.assertEqual(retr_obj.ankle_reflexes_right_foot, PRESENT)
        self.assertEqual(retr_obj.monofilament_left_foot, NORMAL)
        self.assertEqual(retr_obj.calculated_patient_history_score, 0)
        self.assertEqual(retr_obj.calculated_physical_assessment_score, 0)

    def test_mnsi_calculations_saved_with_mnsi_not_performed_responses(self):
        responses = self.get_mnsi_not_performed_answers()
        model = self.get_mnsi_obj(**responses)

        # Confirm some assumptions about model attributes
        # (both non-calculated and calculated)
        self.assertEqual(model.mnsi_performed, NO)
        self.assertEqual(model.mnsi_not_performed_reason, "e.g. right foot in bandage")
        self.assertEqual(model.numb_legs_feet, NOT_APPLICABLE)
        self.assertEqual(model.ankle_reflexes_right_foot, NOT_EXAMINED)
        self.assertEqual(model.monofilament_left_foot, NOT_EXAMINED)
        self.assertEqual(model.calculated_patient_history_score, 0)
        self.assertEqual(model.calculated_physical_assessment_score, 0)

        # Retrieve saved object, and confirm the same assumptions hold
        retr_obj = Mnsi.objects.get(id=model.id)
        self.assertEqual(retr_obj.mnsi_performed, NO)
        self.assertEqual(retr_obj.mnsi_not_performed_reason, "e.g. right foot in bandage")
        self.assertEqual(retr_obj.numb_legs_feet, NOT_APPLICABLE)
        self.assertEqual(retr_obj.ankle_reflexes_right_foot, NOT_EXAMINED)
        self.assertEqual(retr_obj.monofilament_left_foot, NOT_EXAMINED)
        self.assertEqual(retr_obj.calculated_patient_history_score, 0)
        self.assertEqual(retr_obj.calculated_physical_assessment_score, 0)

    def test_mnsi_calculations_saved_with_worst_case_responses(self):
        responses = self.get_best_case_answers()
        responses.update(self.get_worst_case_patient_history_data())
        responses.update(self.get_worst_case_physical_assessment_data())
        model = self.get_mnsi_obj(**responses)

        # Confirm some assumptions about model attributes
        # (both non-calculated and calculated)
        self.assertEqual(model.numb_legs_feet, YES)
        self.assertEqual(model.ankle_reflexes_right_foot, ABSENT)
        self.assertEqual(model.monofilament_left_foot, ABSENT)
        self.assertEqual(model.calculated_patient_history_score, 13)
        self.assertEqual(model.calculated_physical_assessment_score, 10)

        # Retrieve saved object, and confirm the same assumptions hold
        retr_obj = Mnsi.objects.get(id=model.id)
        self.assertEqual(retr_obj.numb_legs_feet, YES)
        self.assertEqual(retr_obj.ankle_reflexes_right_foot, ABSENT)
        self.assertEqual(retr_obj.monofilament_left_foot, ABSENT)
        self.assertEqual(retr_obj.calculated_patient_history_score, 13)
        self.assertEqual(retr_obj.calculated_physical_assessment_score, 10)



from edc_list_data.model_mixins import ListModelMixin


class AbnormalFootAppearanceObservations(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Abnormal Foot Appearance Observations"
        verbose_name_plural = "Abnormal Foot Appearance Observations"



from django.db import models

from ..calculator import MnsiCalculator


class MnsiMethodsModelMixin(models.Model):
    def update_mnsi_calculated_fields(self) -> None:
        """Calculates the MNSI scores and updates fields.

        Called in a signal.
        """
        mnsi_calculator = MnsiCalculator(model_obj=self)
        self.calculated_patient_history_score = mnsi_calculator.patient_history_score()
        self.calculated_physical_assessment_score = mnsi_calculator.physical_assessment_score()
        self.save(
            update_fields=[
                "calculated_patient_history_score",
                "calculated_physical_assessment_score",
            ]
        )

    class Meta:
        abstract = True



from django import forms
from edc_form_validators import FormValidatorMixin
from edc_sites.forms import SiteModelFormMixin

from .form_validator import MnsiFormValidator
from .models import Mnsi


class MnsiForm(SiteModelFormMixin, FormValidatorMixin, forms.ModelForm):
    form_validator_cls = MnsiFormValidator

    class Meta:
        model = Mnsi
        fields = "__all__"



from django.db import models
from edc_constants.choices import YES_NO_NOT_EXAMINED
from edc_constants.constants import NOT_EXAMINED
from edc_model import models as edc_models

from .choices import (
    ANKLE_REFLEX_CHOICES,
    MONOFILAMENT_CHOICES,
    ULCERATION_CHOICES,
    VIBRATION_PERCEPTION_CHOICES,
)
from .models import AbnormalFootAppearanceObservations


def foot_exam_model_mixin_factory(foot_choice):
    """Returns an abstract model mixin"""

    class AbstractModel(models.Model):
        class Meta:
            abstract = True

    yes_no_not_examined_options = dict(
        max_length=15,
        choices=YES_NO_NOT_EXAMINED,
        default=NOT_EXAMINED,
        help_text="If the MNSI assessment was not performed, respond with `not examined`.",
    )

    not_examined_options = dict(
        max_length=35,
        default=NOT_EXAMINED,
        help_text=(
            "If the MNSI assessment was not performed, or amputation prevents "
            "examination, respond with `not examined`."
        ),
    )

    attrs = {
        f"normal_appearance_{foot_choice}_foot": models.CharField(
            verbose_name=f"Does {foot_choice.upper()} foot appear normal?",
            **yes_no_not_examined_options,
        ),
        f"abnormal_obs_{foot_choice}_foot": models.ManyToManyField(
            # "edc_mnsi.abnormalfootappearanceobservations",
            AbnormalFootAppearanceObservations,
            related_name="+",
            verbose_name=f"If NO, check all that apply to {foot_choice.upper()} foot?",
            blank=True,
        ),
        f"abnormal_obs_{foot_choice}_foot_other": edc_models.OtherCharField(
            verbose_name=(
                "If other abnormality observed on "
                f"{foot_choice.upper()} foot, please specify ..."
            ),
        ),
        f"ulceration_{foot_choice}_foot": models.CharField(
            verbose_name=f"Ulceration, {foot_choice.upper()} foot?",
            choices=ULCERATION_CHOICES,
            **not_examined_options,
        ),
        f"ankle_reflexes_{foot_choice}_foot": models.CharField(
            verbose_name=f"Ankle reflexes, {foot_choice.upper()} foot?",
            choices=ANKLE_REFLEX_CHOICES,
            **not_examined_options,
        ),
        f"vibration_perception_{foot_choice}_toe": models.CharField(
            verbose_name=f"Vibration perception at great toe, {foot_choice.upper()} foot?",
            choices=VIBRATION_PERCEPTION_CHOICES,
            **not_examined_options,
        ),
        f"monofilament_{foot_choice}_foot": models.CharField(
            verbose_name=f"Monofilament, {foot_choice.upper()} foot?",
            choices=MONOFILAMENT_CHOICES,
            **not_examined_options,
        ),
    }
    for name, fld_cls in attrs.items():
        AbstractModel.add_to_class(name, fld_cls)
    return AbstractModel



from edc_constants.constants import (
    ABSENT,
    DECREASED,
    NORMAL,
    NOT_EXAMINED,
    PRESENT,
    PRESENT_WITH_REINFORCEMENT,
    REDUCED,
)

ANKLE_REFLEX_CHOICES = (
    (PRESENT, "Present"),
    (PRESENT_WITH_REINFORCEMENT, "Present/Reinforcement"),
    (ABSENT, "Absent"),
    (NOT_EXAMINED, "Not examined"),
)

MONOFILAMENT_CHOICES = (
    (NORMAL, "Normal"),
    (REDUCED, "Reduced"),
    (ABSENT, "Absent"),
    (NOT_EXAMINED, "Not examined"),
)

ULCERATION_CHOICES = (
    (ABSENT, "Absent"),
    (PRESENT, "Present"),
    (NOT_EXAMINED, "Not examined"),
)

VIBRATION_PERCEPTION_CHOICES = (
    (PRESENT, "Present"),
    (DECREASED, "Decreased"),
    (ABSENT, "Absent"),
    (NOT_EXAMINED, "Not examined"),
)



from django.contrib import admin
from django_audit_fields import audit_fieldset_tuple
from edc_list_data.admin import ListModelAdminMixin
from edc_model_admin.history import SimpleHistoryAdmin

from edc_mnsi.fieldsets import calculated_values_fieldset
from edc_mnsi.fieldsets import get_fieldsets as get_mnsi_fieldsets
from edc_mnsi.model_admin_mixin import MnsiModelAdminMixin
from edc_mnsi.models import AbnormalFootAppearanceObservations

from .admin_site import edc_mnsi_admin
from .forms import MnsiForm
from .models import Mnsi


def get_fieldsets():
    fieldset = (
        None,
        {
            "fields": (
                "subject_identifier",
                "report_datetime",
                "mnsi_performed",
                "mnsi_not_performed_reason",
            )
        },
    )

    fieldsets = (fieldset,) + get_mnsi_fieldsets()
    fieldsets += (calculated_values_fieldset,)
    fieldsets += (audit_fieldset_tuple,)
    return fieldsets


@admin.register(AbnormalFootAppearanceObservations, site=edc_mnsi_admin)
class AbnormalFootAppearanceObservationsAdmin(ListModelAdminMixin, admin.ModelAdmin):
    pass


@admin.register(Mnsi, site=edc_mnsi_admin)
class MnsiAdmin(
    MnsiModelAdminMixin,
    SimpleHistoryAdmin,
):
    form = MnsiForm

    fieldsets = get_fieldsets()



from __future__ import annotations

from typing import TYPE_CHECKING

from edc_constants.constants import (
    ABSENT,
    DECREASED,
    NO,
    PRESENT,
    PRESENT_WITH_REINFORCEMENT,
    REDUCED,
    YES,
)

if TYPE_CHECKING:
    from edc_mnsi.model_mixins import MnsiModelMixin


class MnsiPatientHistoryCalculatorError(Exception):
    pass


class MnsiPhysicalAssessmentCalculatorError(Exception):
    pass


class MnsiCalculator:
    """Michigan Neuropathy Screening Instrument (MNSI) calculators."""

    def __init__(self, model_obj: MnsiModelMixin | None = None, **kwargs):
        if model_obj:
            self.responses = model_obj.__dict__
        else:
            self.responses = kwargs

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(responses={self.responses})"

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(responses={self.responses})"

    def patient_history_score(self) -> int:
        """Returns MNSI score based on patient history questionnaire.

        Scoring based on:
            https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3641573/#S6title
        """
        try:
            # 1 point added for each 'YES' response for following Qs
            score = [
                self.responses["numb_legs_feet"],  # Q1
                self.responses["burning_pain_legs_feet"],  # Q2
                self.responses["feet_sensitive_touch"],  # Q3
                self.responses["prickling_feelings_legs_feet"],  # Q5
                self.responses["covers_touch_skin_painful"],  # Q6
                self.responses["open_sore_foot_history"],  # Q8
                self.responses["diabetic_neuropathy"],  # Q9
                self.responses["symptoms_worse_night"],  # Q11
                self.responses["legs_hurt_when_walk"],  # Q12
                self.responses["skin_cracks_open_feet"],  # Q14
                self.responses["amputation"],  # Q15
            ].count(YES)

            # 1 point added for each 'NO' response for following Qs
            score += [
                self.responses["differentiate_hot_cold_water"],  # Q7
                self.responses["sense_feet_when_walk"],  # Q13
            ].count(NO)

            # Note: Questions 4 (muscle_cramps_legs_feet) and 10 (feel_weak)
            # not included in scoring algorithm

        except KeyError as exc:
            raise MnsiPatientHistoryCalculatorError(
                f"Can't calculate patient history score for MNSI. "
                f"Expected response '{exc.args[0]}' "
                f"was missing from received responses: {self.responses.keys()}. "
                "Perhaps catch this in the form validation."
            )
        return score

    def physical_assessment_score(self) -> float:
        """Returns score based on lower extremity examination.

        Scoring based on:
            https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3641573/#S6title
            (plus up to 2 extra points awarded for monofilament assessments)
        """
        score = 0.0
        try:
            for foot_choice in ["left", "right"]:
                score += self._get_appearance_points(
                    self.responses[f"normal_appearance_{foot_choice}_foot"]
                )
                score += self._get_ulceration_points(
                    self.responses[f"ulceration_{foot_choice}_foot"]
                )
                score += self._get_ankle_reflex_points(
                    self.responses[f"ankle_reflexes_{foot_choice}_foot"]
                )
                score += self._get_vibration_perception_points(
                    self.responses[f"vibration_perception_{foot_choice}_toe"]
                )
                score += self._get_monofilament_points(
                    self.responses[f"monofilament_{foot_choice}_foot"]
                )
        except KeyError as exc:
            raise MnsiPhysicalAssessmentCalculatorError(
                f"Can't calculate physical assessment score for MNSI. "
                f"Expected response '{exc.args[0]}' "
                f"was missing from received responses: {self.responses.keys()}. "
                "Perhaps catch this in the form validation."
            )
        return score

    @staticmethod
    def _get_appearance_points(response: str) -> int:
        return 1 if response == NO else 0

    @staticmethod
    def _get_ulceration_points(response: str) -> int:
        return 1 if response == PRESENT else 0

    @staticmethod
    def _get_ankle_reflex_points(response: str) -> float:
        if response == PRESENT_WITH_REINFORCEMENT:
            return 0.5
        elif response == ABSENT:
            return 1.0
        return 0.0

    @staticmethod
    def _get_vibration_perception_points(response: str) -> float:
        if response == DECREASED:
            return 0.5
        elif response == ABSENT:
            return 1.0
        return 0.0

    @staticmethod
    def _get_monofilament_points(response: str) -> float:
        if response == REDUCED:
            return 0.5
        elif response == ABSENT:
            return 1.0
        return 0.0



from copy import deepcopy

from django.test import TestCase
from edc_constants.constants import (
    ABSENT,
    NO,
    NORMAL,
    NOT_APPLICABLE,
    NOT_EXAMINED,
    OTHER,
    PRESENT,
    YES,
)
from edc_list_data import site_list_data
from edc_utils import get_utcnow

from edc_mnsi import list_data
from edc_mnsi.models import AbnormalFootAppearanceObservations, Mnsi


class TestCaseMixin(TestCase):
    foot_choices = ["right", "left"]

    @classmethod
    def setUpTestData(cls):
        site_list_data.initialize()
        site_list_data.register(list_data, app_name="edc_mnsi")
        site_list_data.load_data()

    def setUp(self):
        self.subject_identifier = "1234"
        self.data = dict(
            subject_identifier=self.subject_identifier, report_datetime=get_utcnow()
        )

    def get_mnsi_obj(
        self, abnormal_obs_left_foot=None, abnormal_obs_right_foot=None, **responses
    ):
        """Returns an Mnsi model instance with the m2ms added"""
        responses.update(self.data)
        mnsi = Mnsi(**responses)
        mnsi.save()
        for obj in abnormal_obs_left_foot:
            mnsi.abnormal_obs_left_foot.add(obj)
        for obj in abnormal_obs_right_foot:
            mnsi.abnormal_obs_right_foot.add(obj)
        return mnsi

    @staticmethod
    def get_empty_abnormal_obs_set():
        return AbnormalFootAppearanceObservations.objects.filter(name=None)

    @staticmethod
    def get_nonempty_abnormal_obs_set(value="infection"):
        return AbnormalFootAppearanceObservations.objects.filter(name=value)

    @property
    def amputated_abnormal_obs_sets(self):
        return [
            self.get_nonempty_abnormal_obs_set("deformity_amputation"),
            self.get_amputation_plus_other_abnormalities_set(),
        ]

    @staticmethod
    def get_amputation_plus_other_abnormalities_set():
        return AbnormalFootAppearanceObservations.objects.exclude(name__contains=OTHER)

    def get_best_case_answers(self):
        data = deepcopy(self.data)
        data.update(
            {
                "mnsi_performed": YES,
                "mnsi_not_performed_reason": None,
            }
        )
        data.update(self.get_best_case_patient_history_data())
        data.update(self.get_best_case_physical_assessment_data())
        return data

    def get_best_case_patient_history_data(self):
        data = deepcopy(self.data)
        # Part 1: Patient History
        data.update(
            {
                "numb_legs_feet": NO,
                "burning_pain_legs_feet": NO,
                "feet_sensitive_touch": NO,
                "muscle_cramps_legs_feet": NO,  # no effect on score, regardless of value
                "prickling_feelings_legs_feet": NO,
                "covers_touch_skin_painful": NO,
                "differentiate_hot_cold_water": YES,
                "open_sore_foot_history": NO,
                "diabetic_neuropathy": NO,
                "feel_weak": NO,  # no effect on score, regardless of value
                "symptoms_worse_night": NO,
                "legs_hurt_when_walk": NO,
                "sense_feet_when_walk": YES,
                "skin_cracks_open_feet": NO,
                "amputation": NO,
            }
        )
        return data

    def get_best_case_physical_assessment_data(self):
        data = deepcopy(self.data)
        for foot_choice in self.foot_choices:
            data.update(
                {
                    f"normal_appearance_{foot_choice}_foot": YES,
                    f"abnormal_obs_{foot_choice}_foot": self.get_empty_abnormal_obs_set(),
                    f"ulceration_{foot_choice}_foot": ABSENT,
                    f"ankle_reflexes_{foot_choice}_foot": PRESENT,
                    f"vibration_perception_{foot_choice}_toe": PRESENT,
                    f"monofilament_{foot_choice}_foot": NORMAL,
                }
            )
        return data

    def get_mnsi_not_performed_answers(self):
        data = deepcopy(self.data)
        data.update(
            {
                "mnsi_performed": NO,
                "mnsi_not_performed_reason": "e.g. right foot in bandage",
            }
        )
        # Part 1: Patient History
        data.update(
            {
                "numb_legs_feet": NOT_APPLICABLE,
                "burning_pain_legs_feet": NOT_APPLICABLE,
                "feet_sensitive_touch": NOT_APPLICABLE,
                "muscle_cramps_legs_feet": NOT_APPLICABLE,
                "prickling_feelings_legs_feet": NOT_APPLICABLE,
                "covers_touch_skin_painful": NOT_APPLICABLE,
                "differentiate_hot_cold_water": NOT_APPLICABLE,
                "open_sore_foot_history": NOT_APPLICABLE,
                "diabetic_neuropathy": NOT_APPLICABLE,
                "feel_weak": NOT_APPLICABLE,
                "symptoms_worse_night": NOT_APPLICABLE,
                "legs_hurt_when_walk": NOT_APPLICABLE,
                "sense_feet_when_walk": NOT_APPLICABLE,
                "skin_cracks_open_feet": NOT_APPLICABLE,
                "amputation": NOT_APPLICABLE,
            }
        )
        for foot_choice in self.foot_choices:
            data.update(
                {
                    f"normal_appearance_{foot_choice}_foot": NOT_EXAMINED,
                    f"abnormal_obs_{foot_choice}_foot": self.get_empty_abnormal_obs_set(),
                    f"ulceration_{foot_choice}_foot": NOT_EXAMINED,
                    f"ankle_reflexes_{foot_choice}_foot": NOT_EXAMINED,
                    f"vibration_perception_{foot_choice}_toe": NOT_EXAMINED,
                    f"monofilament_{foot_choice}_foot": NOT_EXAMINED,
                }
            )
        return data

    def get_worst_case_patient_history_data(self):
        data = deepcopy(self.data)
        # Part 1: Patient History
        data.update(
            {
                "numb_legs_feet": YES,
                "burning_pain_legs_feet": YES,
                "feet_sensitive_touch": YES,
                "prickling_feelings_legs_feet": YES,
                "covers_touch_skin_painful": YES,
                "differentiate_hot_cold_water": NO,
                "open_sore_foot_history": YES,
                "diabetic_neuropathy": YES,
                "symptoms_worse_night": YES,
                "legs_hurt_when_walk": YES,
                "sense_feet_when_walk": NO,
                "skin_cracks_open_feet": YES,
                "amputation": YES,
            }
        )
        return data

    def get_worst_case_physical_assessment_data(self):
        data = deepcopy(self.data)
        for foot_choice in self.foot_choices:
            data.update(
                {
                    f"normal_appearance_{foot_choice}_foot": NO,
                    f"abnormal_obs_{foot_choice}_foot": self.get_nonempty_abnormal_obs_set(),
                    f"ulceration_{foot_choice}_foot": PRESENT,
                    f"ankle_reflexes_{foot_choice}_foot": ABSENT,
                    f"vibration_perception_{foot_choice}_toe": ABSENT,
                    f"monofilament_{foot_choice}_foot": ABSENT,
                }
            )
        return data



from edc_constants.constants import NO, NOT_EXAMINED, OTHER, YES
from edc_crf.crf_form_validator import CrfFormValidator

from .calculator import MnsiCalculator
from .fieldsets import patient_history_fields


class MnsiFormValidatorMixin:
    def clean(self):
        self.required_if(
            NO,
            field="mnsi_performed",
            field_required="mnsi_not_performed_reason",
        )
        self.clean_patient_history_fields()
        self.clean_physical_assessments()
        self.clean_calculation_data()

    def clean_patient_history_fields(self) -> None:
        for field in patient_history_fields:
            self.applicable_if(
                YES,
                field="mnsi_performed",
                field_applicable=field,
            )

    def foot_amputated(self, foot_choice: str) -> bool:
        if self.cleaned_data.get(
            f"normal_appearance_{foot_choice}_foot"
        ) == NO and self.cleaned_data.get(f"abnormal_obs_{foot_choice}_foot"):
            qs = self.cleaned_data.get(f"abnormal_obs_{foot_choice}_foot")
            return qs.filter(name="deformity_amputation").exists()
        return False

    def clean_physical_assessments(self) -> None:
        applicable_if_opts = dict(
            not_applicable_value=NOT_EXAMINED,
            applicable_msg=(
                "Invalid. Examination result expected if MNSI assessment was performed."
            ),
            not_applicable_msg=(
                "Invalid. Expected `not examined` if MNSI assessment was not performed.",
            ),
        )
        for foot_choice in ["right", "left"]:
            self.applicable_if(
                YES,
                field="mnsi_performed",
                field_applicable=f"normal_appearance_{foot_choice}_foot",
                **applicable_if_opts,
            )

            self.m2m_required_if(
                response=NO,
                field=f"normal_appearance_{foot_choice}_foot",
                m2m_field=f"abnormal_obs_{foot_choice}_foot",
            )

            self.m2m_other_specify(
                OTHER,
                m2m_field=f"abnormal_obs_{foot_choice}_foot",
                field_other=f"abnormal_obs_{foot_choice}_foot_other",
            )

            for target_field in [
                f"ulceration_{foot_choice}_foot",
                f"ankle_reflexes_{foot_choice}_foot",
                f"vibration_perception_{foot_choice}_toe",
                f"monofilament_{foot_choice}_foot",
            ]:
                if not self.foot_amputated(foot_choice):
                    self.applicable_if(
                        YES,
                        field="mnsi_performed",
                        field_applicable=target_field,
                        **applicable_if_opts,
                    )

    def clean_calculation_data(self) -> None:
        mnsi_calculator = MnsiCalculator(**self.cleaned_data)
        mnsi_calculator.patient_history_score()
        mnsi_calculator.physical_assessment_score()


class MnsiFormValidator(MnsiFormValidatorMixin, CrfFormValidator):
    pass



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from edc_constants.choices import YES_NO, YES_NO_NA
from edc_constants.constants import NOT_APPLICABLE, YES

from ..factory import foot_exam_model_mixin_factory


class MnsiRightFootFieldsModelMixin(foot_exam_model_mixin_factory("right"), models.Model):
    class Meta:
        abstract = True


class MnsiLeftFootFieldsModelMixin(foot_exam_model_mixin_factory("left"), models.Model):
    class Meta:
        abstract = True


class MnsiFieldsModelMixin(models.Model):
    mnsi_performed = models.CharField(
        verbose_name="Is the MNSI assessment being performed?",
        max_length=15,
        choices=YES_NO,
        default=YES,
        help_text=(
            "If completion of patient history or physical assessment not possible, "
            "respond with `no` and provide reason below."
        ),
    )

    mnsi_not_performed_reason = models.TextField(
        verbose_name="If NO, please provide a reason",
        max_length=200,
        null=True,
        blank=True,
    )

    numb_legs_feet = models.CharField(
        verbose_name="Are your legs and/or feet numb?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    burning_pain_legs_feet = models.CharField(
        verbose_name="Do you ever have any burning pain in your legs and/or feet?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    feet_sensitive_touch = models.CharField(
        verbose_name="Are your feet too sensitive to touch?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    muscle_cramps_legs_feet = models.CharField(
        verbose_name="Do you get muscle cramps in your legs and/or feet?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    prickling_feelings_legs_feet = models.CharField(
        verbose_name="Do you ever have any prickling feelings in your legs or feet?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    covers_touch_skin_painful = models.CharField(
        verbose_name="Does it hurt when the bed covers touch your skin?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    differentiate_hot_cold_water = models.CharField(
        verbose_name=(
            "When you get into the tub or shower, are you able to tell the hot water from the "
            "cold water?"
        ),
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    open_sore_foot_history = models.CharField(
        verbose_name="Have you ever had an open sore on your foot?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    diabetic_neuropathy = models.CharField(
        verbose_name="Has your doctor ever told you that you have diabetic neuropathy?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    feel_weak = models.CharField(
        verbose_name="Do you feel weak all over most of the time?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    symptoms_worse_night = models.CharField(
        verbose_name="Are your symptoms worse at night?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    legs_hurt_when_walk = models.CharField(
        verbose_name="Do your legs hurt when you walk?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    sense_feet_when_walk = models.CharField(
        verbose_name="Are you able to sense your feet when you walk?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    skin_cracks_open_feet = models.CharField(
        verbose_name="Is the skin on your feet so dry that it cracks open?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    amputation = models.CharField(
        verbose_name="Have you ever had an amputation?",
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    calculated_patient_history_score = models.IntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(13)],
        null=True,
        blank=True,
    )

    calculated_physical_assessment_score = models.DecimalField(
        max_digits=3,
        decimal_places=1,
        validators=[MinValueValidator(0), MaxValueValidator(10.0)],
        null=True,
        blank=True,
    )

    class Meta:
        abstract = True
